#! /usr/bin/python3 -O
# -*- coding: utf-8 -*-
#
# Prime numbers.
# Copyright (C) 2022 MiKael NAVARRO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Prime numbers.
Copyright (C) 2022 MiKael NAVARRO

Find prime numbers less than or equal to {num}.

A prime number (or a prime) is a natural number greater than 1
that is not a product of two smaller natural numbers.

"""

# Specific variables for pydoc
__author__ = "MiKael Navarro <klnavajo@gmail.com>"
__date__ = "Fri Sep 16 2022"
__version__ = "0.0.4"
__credits__ = """https://www.dcode.fr/recherche-nombres-premiers"""

# Include directives
import os, sys
import math

from time import perf_counter, process_time
from timeit import Timer
from datetime import timedelta

from multiprocessing import Pool

HAVE_JOBLIB = True
try:
    import joblib
except ImportError:
    HAVE_JOBLIB = False

HAVE_NUMBA = True
try:
    from numba import njit, prange
except ImportError:
    HAVE_NUMBA = False

HAVE_NUMPY = True
try:
    import numpy as np
except ImportError:
    HAVE_NUMPY = False

HAVE_BITARRAY = True
try:
    from bitarray import bitarray
except ImportError:
    HAVE_BITARRAY = False

# Global variables
METHOD = [ "simple", "multi" ]
if HAVE_JOBLIB:
    METHOD += [ "joblib" ]
METHOD += [ "crible" ]
    
PROCESSES = len(os.sched_getaffinity(0))  #os.cpu_count()


#
# Primality test:
# Cf. https://en.wikipedia.org/wiki/Primality_test
#
def is_prime(num):
    """Determining whether an input {num} is prime.  Return the prime number itself on None.

    Using the simplest trial division test:
    Check whether {num} is evenly divisible by one of the integers between 2 and sqrt(num).

    >>> is_prime(97)
    97
    >>> is_prime(100)
    
    """

    assert type(num) is int, f"{num} is not an integer!"
    assert num > 0, f"{num} is not a natural number (positive integers)!"
    
    if __debug__: print(f"{num}/2 = {divmod(num, 2)}")
    if (num % 2 == 0):
        return None

    for n in range(3, int(math.sqrt(num))+1, 2):  # iterator
        if __debug__: print(f"{num}/{n} = {divmod(num, n)}")
        if (num % n == 0):
            return None
    else:
        return num

if HAVE_NUMBA:
    # Using the Numba JIT compilation.
    # Cf. http://numba.pydata.org/numba-doc/latest/user/parallel.html#explicit-parallel-loops
    @njit
    def is_prime_jit(num):
        if (num % 2 == 0):
            return None
        for n in prange(3, int(math.sqrt(num))+1, 2):
            if (num % n == 0):
                return None
        else:
            return num

    # Using 6k +/- 1 optimization.
    @njit
    def is_prime_optim_jit(num):
        if num <= 3:
            return (num > 1) and num or None
        if not num % 2 or not num % 3:
            return None
        for n in prange(5, int(num**0.5)+1, 6):
            if not num % n or not num % (n+2):
                return None
        else:
            return num


#
# Find prime numbers <= {num}:
#
def prime_numbers(num, use_numba=False):
    """Using a naïve method of determining prime numbers: primality test.

    >>> prime_numbers(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
 
    """

    res = [ 2 ]  # init list with the prime number 2
    for n in range(3, num, 2):  # test only odd number
        if HAVE_NUMBA and use_numba:
            if is_prime_optim_jit(n):
                res.append(n)
        else:
            if is_prime(n):
                res.append(n)

    return res


#
# Find prime numbers <= {num} using the multiprocessing module to parallelize the for loop.
#
def prime_numbers_multi(num, chunksize=1000, use_numba=False):
    """Using the multiprocessing module to parallelize tasks:

    >>> prime_numbers_multi(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
 
    """

    with Pool(PROCESSES) as p:
        if HAVE_NUMBA and use_numba:
            res = p.imap(is_prime_optim_jit, prange(3, num, 2), chunksize=chunksize)
        else:
            res = p.imap(is_prime, range(3, num, 2), chunksize=chunksize)
        return [ 2 ] + [ n for n in res if n ]

# Using an iterator to return prime numbers
def prime_numbers_iter_multi(num, chunksize=1000000, use_numba=False):
    """Using the multiprocessing module to parallelize tasks:

    >>> prime_numbers_multi(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
 
    """

    yield 2  # first prime number
    with Pool(PROCESSES) as p:
        if HAVE_NUMBA and use_numba:
            for p in p.imap(is_prime_optim_jit, prange(3, num, 2), chunksize=chunksize):
                if p:
                    yield p
        else:
            for p in p.imap(is_prime, prange(3, num, 2), chunksize=chunksize):
                if p:
                    yield p


if HAVE_JOBLIB:
    # Find prime numbers <= {num} using the joblib module to parallelize the for loop.
    # Cf. https://joblib.readthedocs.io/en/latest/
    #
    def prime_numbers_joblib(num, jobs=2, use_numba=False):
        """Using the joblib module to parallelize tasks:
        
        >>> prime_numbers_joblib(100)
        [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
        
        """
    
        if HAVE_NUMBA and use_numba:
            res = joblib.Parallel(n_jobs=jobs, prefer='processes')(joblib.delayed(is_prime_optim_jit)(n) for n in prange(3, num, 2))
        else:
            res = joblib.Parallel(n_jobs=jobs, prefer='processes')(joblib.delayed(is_prime)(n) for n in range(3, num, 2))
        return [ 2 ] + [ n for n in res if n ]


#
# Find prime numbers <= {num} using the sieve of Eratosthenes.
# Cf. https://en.wikipedia.org/wiki/Sieve_of_Eratosthenes
#
def eratosthene(num):
    """Using the sieve of Eratosthene:

    >>> eratosthene(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
 
    """

    numbers = [ True ] * (num+1)  # => [0..num]
    numbers[0] = False  # out of scope
    numbers[1] = False
    
    for n in range(2, num+1):  # iterator
        if numbers[n]:
            m = 2
            while True:
                try:
                    numbers[n*m] = False  # remove all multiples
                    m += 1
                except IndexError:
                    break
    
    return [ n for n,f in enumerate(numbers) if f ]

if HAVE_NUMBA:
    # Using the Numba JIT compilation.
    @njit
    def eratosthene_jit(num):
        tab = [ False, False ] + [ True ] * (num-2)  # => [0..num]
        tab[2::2] = [ False ] * ((num-2)//2 + num%2)  # supp even numbers
        
        primes = [ 2 ]  # init list of prime numbers
        
        sqrt = int(num**0.5)
        sqrt = sqrt + [1,0][sqrt%2]  # next odd number
        
        for n in prange(3, sqrt+1, 2):
            if tab[n]:
                primes.append(n)
                tab[n::n] = [ False ] * ((num-n)//n + int((num-n)%n>0))  # remove all multiples

        for n in prange(sqrt, num, 2):
            if tab[n]:
                primes.append(n)
                
        return primes

    # Using an iterator to return prime numbers
    @njit
    def eratosthene_iter_jit(num):
        tab = [ False, False ] + [ True ] * (num-2)  # => [0..num]
        tab[2::2] = [ False ] * ((num-2)//2 + num%2)  # supp even numbers
        
        yield 2  # 2 is a prime numbers
        
        sqrt = int(num**0.5)
        sqrt = sqrt + [1,0][sqrt%2]  # next odd number

        for n in prange(3, sqrt+1, 2):
            if tab[n]:
                yield n  # n is a prime number
                tab[n::n] = [ False ] * ((num-n)//n + int((num-n)%n>0))  # remove all multiples

        for n in prange(sqrt, num, 2):
            if tab[n]:
                yield n  # n is a prime number

if HAVE_BITARRAY:
    # Using bitarray + an iterator to return prime numbers
    def eratosthene_iter_bitarray(num):
        tab = bitarray(num+1)
        tab.setall(True)  # => [0..num]
        tab[0], tab[1] = False, False        
        tab[2::2] = False
        
        yield 2  # 2 is a prime numbers
        
        sqrt = int(num**0.5)
        sqrt = sqrt + [1,0][sqrt%2]  # next odd number

        for n in range(3, sqrt+1, 2):
            if tab[n]:
                yield n  # n is a prime number
                tab[n::n] = False

        for n in range(sqrt, num, 2):
            if tab[n]:
                yield n  # n is a prime number

if HAVE_NUMBA and HAVE_NUMPY:
    # Using NumPy array + an iterator to return prime numbers
    @njit
    def eratosthene_iter2_jit(num):
        tab = np.arange(num+1).astype(np.bool_)  # => [0..num]
        tab[0], tab[1] = False, False        
        tab[2::2] = False
        
        yield 2  # 2 is a prime numbers
        
        sqrt = int(num**0.5)
        sqrt = sqrt + [1,0][sqrt%2]  # next odd number

        for n in prange(3, sqrt+1, 2):
            if tab[n]:
                yield n  # n is a prime number
                tab[n::n] = False

        for n in prange(sqrt, num, 2):
            if tab[n]:
                yield n  # n is a prime number
    

#
# Utility function to print sequence:
#
def disp(sequence, limit=1000):
    """Display {sequence} like a list."""

    print("[", end="")
    for c,n in enumerate(sequence, start=1):
        if c <= limit:
            print(n, end=", ")
    if c > limit:
        print("...", end=", ")
    print("\b\b]", c)


#
# Tests.
#
def test():
    """Doc tests."""
    
    import doctest
    doctest.testmod()
    

#
# Main entry point.
#
def main(args):
    """Find prime numbers using different methods."""

    if args.test:
        test()

    if args.perf:
        t0, t1 = perf_counter(), process_time()

    # Find prime numbers:
    print("Find prime numbers less than or equal to {:,}:".format(args.num).replace(',', ' '))
    
    if args.method == "multi":
        print("Using \"naïve\" method + multiprocessing{}".format(" (with Numba JIT compilation)..." if (HAVE_NUMBA and args.numba) else "..."))
        disp(prime_numbers_iter_multi(args.num, args.chunksize, args.numba))

    elif HAVE_JOBLIB and args.method == "joblib":
        print("Using \"naïve\" method + joblib{}".format(" (with Numba JIT compilation)..." if (HAVE_NUMBA and args.numba) else "..."))
        disp(prime_numbers_joblib(args.num, args.jobs, args.numba))

    elif args.method == "crible":
        print("Using the sieve of Eratosthene", end="")

        if HAVE_BITARRAY and not args.numba:
            print(" + bitarray...")
            disp(eratosthene_iter_bitarray(args.num))  # => 50 847 533 / 31" (1.08G/222M) ; 455 052 511 / 5'25" (2.86G/1.26G) ; 4 118 054 813 / 55’ (13.3G/11.7G) ; 37 607 912 018 / MemoryError!
        else:
            if HAVE_NUMBA and args.numba:
                if HAVE_NUMPY:
                    print(" + NumPy (with Numba JIT compilation)...")
                    disp(eratosthene_iter2_jit(args.num))  # => 50 847 533 / 18" (2.65G/1.05G) ; 455 052 511 / segmentation fault (core dumped)!!!
                else:
                    print(" (with Numba JIT compilation)...")
                    #disp(eratosthene_jit(args.num))  # => 50 847 533 / 22" (2.67G/1.08G)
                    disp(eratosthene_iter_jit(args.num))  # => 50 847 533 / 22" (2.67G/1.08G) ; 455 052 511 / 16' (20.3G/10.3G)
            else:
                print("...")
                disp(eratosthene(args.num))  # if not args.numba and not HAVE_BITARRAY
                            
    else:  # "simple"
        print("Using \"naïve\" method{}".format(" (with Numba JIT compilation)..." if (HAVE_NUMBA and args.numba) else "..."))
        disp(prime_numbers(args.num, args.numba))
            
    if args.perf:
        t2, t3 = perf_counter(), process_time()
        elapsed_t = timedelta(seconds=t2-t0)
        process_t = timedelta(seconds=t3-t1)
        print(f"Elapsed time= {elapsed_t} ; Process time= {process_t}")

        # Dell G5 5587:
        #   i7-8750H CPU @ 2.20GHz x 12 ; 16Gio Mem
        #   Ubuntu 22.04 LTS / Python 3.10 (CPython)
        
        # | method         | 1 000 000 | jit | 10 000 000 | jit | 100 000 000 | jit   |
        # |----------------+-----------+-----+------------+-----+-------------+-------|
        # | simple         | 1.67"     |     | 49"        | 5"  | 21'         | 1'36" |
        # | multi(10)      | 4.42"     |     |            |     |             |       |
        # | multi(100)     | 0.59"     |     |            |     |             |       |
        # | multi(1000)    | 0.31"     |     |            |     | 6'          | 2'    |
        # | multi(10000)   | 0.28"     |     |            |     | 6'          | 44"   |
        # | multi(100000)  | 0.14"     |     | 11"        | 2"  | 6'          | 19"   |
        # | multi(1000000) | 0.13"     |     |            |     | 6'          | 14"   |
        # | joblib(10)     | 3"        |     | 20"        | 18" | 6'          | 4'    |
        # | crible         | 0.33"     |     | 4"         |     | 41"         | 5"    |


#
# External entry point.
#
if __name__ == "__main__":
    # Parsing command line (http://www.doughellmann.com/PyMOTW/argparse/)
    import argparse

    parser = argparse.ArgumentParser(description="List prime numbers less than or equal to {num}.",
                                     prefix_chars='-+')
    parser.add_argument('-V', '--version',
                        action='version', version=__version__)
    parser.add_argument('+t', '--test', action="store_true",
                       dest='test', default=None,
                       help="auto tests")
    parser.add_argument('+p', '--perf', action="store_true",
                       dest='perf', default=None,
                       help="measure perfs")
    parser.add_argument('-m', '--method', action='store',
                        dest='method', default="simple",
                        help=f"method to use {METHOD}")
    parser.add_argument('-c', '--chunksize', action='store',
                        type=int, dest='chunksize', default=100,
                        help=f"chunksize for muliprocessing")
    if HAVE_JOBLIB:
        parser.add_argument('-j', '--job', action='store',
                            type=int, dest='jobs', default=2,
                            help=f"number of subprocess for joblib")
    if HAVE_NUMBA:
        parser.add_argument('-n', '--numba', action='store_true',
                            dest='numba', default=False,
                            help=f"use Numba JIT compilation")

    # Arguments
    parser.add_argument('num', action='store',
                        type=int,
                        help="max number (int)")

    # Parse!
    args = parser.parse_args()

    # Process start here
    main(args)
